import hashlib
import json
import os
import re
import requests
import time
from urllib.parse import urlparse, urljoin

from charms.reactive import hook, when, when_not, remove_state, set_state
from charms.reactive.helpers import data_changed
from charmhelpers.core import host, hookenv, unitdata
from charmhelpers.contrib.charmsupport import nrpe


API_PORT = '9001'  # This is the default set by the snap
API_URL = 'http://0.0.0.0:9001/api/'  # This is the default set by the snap
CONF_FILE = '/var/snap/graylog/common/server.conf'
ELASTICSEARCH_DISCOVERY_PORT = '9300'
SERVICE_NAME = 'snap.graylog.graylog'


@hook('upgrade-charm')
def upgrade_charm():
    remove_state('graylog.configured')


@when('config.changed')
def update_config():
    remove_state('graylog.configured')


def api_request(url, auth, method='GET', data=None):
    if method not in ('GET', 'POST', 'PUT'):
        return
    tries = 0
    while tries < 3:
        tries += 1
        try:
            resp = requests.api.request(method, url, auth=auth, data=data,
                                        headers={'Accept': 'application/json',
                                                 'Content-Type': 'application/json'})
            if resp.ok:
                return json.loads(resp.content.decode('utf-8'))
        except Exception:
            pass


def api_retrieve_token(api_base_url, auth, name='graylog-charm'):
    "Retrieve an existing token to use otherwise create a new one"
    url = urljoin(api_base_url, 'users/{}/tokens/'.format(auth[0]))
    r = api_request(url, auth=auth)
    if not r:
        return
    for token in r['tokens']:
        if token['name'] == name:
            return token['token']
    # None found, let's create a new
    url = urljoin(url, name)
    r = api_request(url, auth=auth, method='POST')
    return api_retrieve_token(api_base_url, auth, name=name)


def api_retrieve_index_sets(api_base_url, auth):
    url = urljoin(api_base_url, 'system/indices/index_sets')
    return api_request(url, auth=auth)['index_sets']


rotation_strategies = {
    'time': {
        'class': 'org.graylog2.indexer.rotation.strategies.TimeBasedRotationStrategy',
        'special': 'rotation_period',
        'juju-conf': 'index_rotation_period',
    },
    'size': {
        'class': 'org.graylog2.indexer.rotation.strategies.SizeBasedRotationStrategy',
        'special': 'max_size',
        'juju-conf': 'index_rotation_size',
    },
    'msg_count': {
        'class': 'org.graylog2.indexer.rotation.strategies.MessageCountRotationStrategy',
        'special': 'max_docs_per_index',
        'juju-conf': 'index_rotation_msg_count',
    }
}


@when('snap.installed.graylog')
@when_not('graylog.configured')
def configure_graylog():
    db = unitdata.kv()
    if not os.path.exists(CONF_FILE):
        hookenv.log('Configuration file "{}" missing, skipping configuration run.'.format(CONF_FILE))
        return
    conf = hookenv.config()
    if conf['admin_password']:
        admin_password = conf['admin_password']
    else:
        admin_password = db.get('admin_password', host.pwgen(18))
    db.set('admin_password', admin_password)
    pw_hash = hashlib.sha256(admin_password.encode('utf-8')).hexdigest()
    set_conf('root_password_sha2', pw_hash)
    if conf['web_listen_uri']:
        set_conf('web_listen_uri', conf['web_listen_uri'])

    remove_state('graylog_api.configured')
    set_state('graylog.configured')
    hookenv.status_set('active', 'Ready')


@when('graylog.configured')
@when('mongodb.available')
@when('elasticsearch.available')
@when_not('graylog.needs_restart')
@when_not('graylog_api.configured')
def configure_graylog_via_api(*discard):
    conf = hookenv.config()
    db = unitdata.kv()
    admin_password = db.get('admin_password')
    api_token = api_retrieve_token(API_URL, auth=('admin', admin_password))
    if not api_token:
        return
    index_sets = api_retrieve_index_sets(API_URL, auth=(api_token, 'token'))
    if not index_sets:
        return
    for iset in index_sets:
        iset['shards'] = conf['index_replicas']
        iset['replicas'] = conf['index_shards']
        if conf['index_rotation_strategy'] in rotation_strategies:
            rs = rotation_strategies[conf['index_rotation_strategy']]
            iset['rotation_strategy_class'] = rs['class']
            iset['rotation_strategy'] = {
                'type': '{}Config'.format(rs['class']),
                rs['special']: conf[rs['juju-conf']],
            }
        iset['retention_strategy_class'] = 'org.graylog2.indexer.retention.strategies.DeletionRetentionStrategy'
        iset['retention_strategy'] = {
            'type': 'org.graylog2.indexer.retention.strategies.DeletionRetentionStrategyConfig',
            'max_number_of_indices': conf['index_retention_count'],
        }
        url = urljoin(API_URL, 'system/indices/index_sets/{}'.format(iset['id']))
        if not api_request(url, auth=(api_token, 'token'), method='PUT',
                           data=json.dumps(iset, indent=True)):
            return

    set_state('graylog_api.configured')


@when('graylog.configured')
@when('elasticsearch.available')
def configure_elasticsearch_connection(elasticsearch):
    cluster = ""
    hosts = []
    for unit in elasticsearch.list_unit_data():
        cluster_name = unit['cluster_name']
        if cluster_name is not None:
            cluster = cluster_name
        # The relation is reporting the http port not the zen discovery port
        # hosts.append('{}:{}'.format(unit['host'], unit['port']))
        hosts.append('{}:{}'.format(unit['host'], ELASTICSEARCH_DISCOVERY_PORT))

    if not cluster:
        conf = hookenv.config()
        cluster = conf['elasticsearch_cluster_name']

    if not data_changed('elasticsearch.relation', {'cluster_name': cluster, 'hosts': hosts}):
        return

    set_conf('elasticsearch_cluster_name', cluster)
    set_conf('elasticsearch_discovery_zen_ping_unicast_hosts ', ', '.join(hosts))
    # Elastic search does not reliably pick the right ip to listen on
    set_conf('elasticsearch_network_host', hookenv.unit_private_ip())

    hookenv.log('Updated elasticsearch configuration')
    hookenv.status_set('active', 'Ready')


@when('graylog.configured')
@when('mongodb.available')
def configure_mongodb_connection(mongodb):
    mongodb_hosts = []
    for mongo_host in mongodb.connection_strings():
        mongodb_hosts.append(mongo_host)

    mongodb_uri = 'mongodb://{}/graylog'.format(','.join(mongodb_hosts))
    if not data_changed('mongodb.uri', mongodb_uri):
        return

    set_conf('mongodb_uri', mongodb_uri)
    hookenv.log('Updated mongodb configuration')
    hookenv.status_set('active', 'Ready')


@when('graylog.configured')
@when('website.available')
def update_reverseproxy_config(website):
    services = ""
    conf = hookenv.config()
    if conf['web_listen_uri']:
        url = urlparse(conf['web_listen_uri'])
        website.configure(port=url.port)
        services += "- {service_name: web, service_port: " + str(url.port) + "}\n"

    services += "- {service_name: api, service_port: " + API_PORT + "}\n"
    website.set_remote(all_services=services)


@when('graylog.configured')
@when('graylog.needs_restart')
def restart_service(service=SERVICE_NAME):
    """ Restart the service and on failure wait 15 seconds and try again.
        This handles situations when relations are adding quick enough to trigger
        systemd's standard protection against frequent restarts.
    """
    host.service_restart(service)
    remove_state('graylog.needs_restart')
    if host.service_running(service):
        return

    time.sleep(15)
    host.service_restart(service)


def set_conf(key, value, conf_path=CONF_FILE):
    """ Updates the graylog configuration setting lines matching 'key = value'.
        The configuration contains randomly generated values specific to each install
        so setting the configuration this way avoids having to regenerate those values.
    """
    conf = []
    key_re = re.compile('^#*\s*{}'.format(key))

    changed = False
    replaced = False
    with open(conf_path, 'rb') as conf_file:
        for line in conf_file:
            line = line.decode().rstrip()
            if not key_re.match(line):
                conf.append(line)
                continue

            if not replaced:
                new = '{} = {}'.format(key, value)
                if line != new:
                    hookenv.log('Updating configuration file option "{}".'.format(key))
                    changed = True
                conf.append(new)
                replaced = True
            elif not line.startswith('#'):
                conf.append('#' + line)
            else:
                conf.append(line)

    if changed:
        set_state('graylog.needs_restart')
        with open(conf_path, 'wb') as conf_file:
            conf_file.write('\n'.join(conf).encode())
        return True
    return False


@when('graylog.configured')
@when('nrpe-external-master.available')
def configure_nagios(nagios):
    if hookenv.hook_name() == 'update-status':
        return

    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=True)

    conf = hookenv.config()
    if conf['web_listen_uri']:
        url = urlparse(conf['web_listen_uri'])
        # LP:#1706265: Work around this bug by escaping some characters.
        check_string = '\<title\>Graylog\ Web\ Interface\</title\>'
        nrpe_setup.add_check(
            'graylog_http',
            'Graylog Web UI check',
            '/usr/lib/nagios/plugins/check_http -I {} -p {} -s "{}"'.format(url.hostname, str(url.port), check_string)
        )
    check_string = 'cluster_id'
    nrpe_setup.add_check(
        'graylog_api',
        'Greylog API check',
        '/usr/lib/nagios/plugins/check_http -I 127.0.0.1 -p {} -u /api -s "{}"'.format(API_PORT, check_string)
    )
    nrpe_setup.write()
