#!/usr/bin/python3

import amulet
import unittest
import yaml

from graylogapi import GraylogAPI


ADMIN_PASSWORD = 'admin'
ADMIN_PASSWORD_HASH = '8c6976e5b5410415bde908bd4dee15dfb167a9c873fc4bb8a81f6f2ab448a918'
CLUSTER_NAME = 'amulet'
DEFAULT_API_PORT = '9001'
DEFAULT_WEB_PORT = '9000'


class TestCharm(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.d = amulet.Deployment(series='xenial')

        cls.d.add('graylog')
        cls.d.configure('graylog', {
            'admin_password': ADMIN_PASSWORD,
            'elasticsearch_cluster_name': CLUSTER_NAME,
        })

        cls.d.add('mongodb', 'cs:mongodb')
        cls.d.configure('mongodb', {
            'replicaset': CLUSTER_NAME,
        })
        cls.d.relate('graylog:mongodb', 'mongodb:database')

        cls.d.add('elasticsearch', 'cs:~ivoks/elasticsearch')
        cls.d.configure('elasticsearch', {
            'apt-repository': 'deb https://packages.elastic.co/elasticsearch/2.x/debian stable main',
            'cluster-name': CLUSTER_NAME,
            'firewall_enabled': 'false',
        })
        cls.d.relate('graylog:elasticsearch', 'elasticsearch:client')

        cls.d.add('apache2', 'cs:apache2')
        cls.d.relate('graylog:website', 'apache2:reverseproxy')
        cls.d.expose('apache2')

        cls.d.setup(timeout=900)
        cls.d.sentry.wait()

        cls.elasticsearch = cls.d.sentry['elasticsearch'][0]
        cls.mongodb = cls.d.sentry['mongodb'][0]
        cls.graylog = cls.d.sentry['graylog'][0]

        cls.config = cls.graylog.file_contents('/var/snap/graylog/common/server.conf')
        api_url = "http://{}:9001/api".format(cls.graylog.info['public-address'])
        cls.api = GraylogAPI(api_url, 'admin', ADMIN_PASSWORD)

    def test_configuration_set(self):
        self.assertTrue(ADMIN_PASSWORD_HASH in self.config)

    def test_elasticsearch_active(self):
        elasticsearch_ip = self.elasticsearch.relation('client',
                                                       'graylog:elasticsearch')['private-address']
        self.assertTrue(elasticsearch_ip in self.config)
        resp = self.api.get('/system/indexer/cluster/health')
        self.assertTrue(resp.json()['status'] == 'green')

    def test_mongodb_active(self):
        mongodb_ip = self.mongodb.relation('database', 'graylog:mongodb')['private-address']
        self.assertTrue(mongodb_ip in self.config)
        resp = self.api.get('/cluster')
        self.assertTrue(list(resp.json().values())[0]['is_processing'])

    def test_website_active(self):
        relation = self.graylog.relation('website', 'apache2:reverseproxy')
        self.assertEqual(relation['port'], DEFAULT_WEB_PORT)

        service_ports = {}
        for service in yaml.safe_load(relation['all_services']):
            service_ports[service['service_name']] = service['service_port']

        self.assertEqual(str(service_ports['web']), DEFAULT_WEB_PORT)
        self.assertEqual(str(service_ports['api']), DEFAULT_API_PORT)


if __name__ == "__main__":
    unittest.main()
